#include <stdio.h>
#include <string.h>
#include "GSHandler.h"
#include "PS2VM.h"
#include "INTC.h"
#include "PtrMacro.h"

#define R_REG(a, v, r)					\
	if((a) & 0x4)						\
	{									\
		v = (uint32)(r >> 32);			\
	}									\
	else								\
	{									\
		v = (uint32)(r & 0xFFFFFFFF);	\
	}

#define W_REG(a, v, r)				\
	if((a) & 0x4)					\
	{								\
		(r) &= 0x00000000FFFFFFFF;	\
		(r) |= (uint64)(v) << 32;	\
	}								\
	else							\
	{								\
		(r) &= 0xFFFFFFFF00000000;	\
		(r) |= (v);					\
	}

using namespace Framework;

CGSHandler::CGSHandler()
{
	m_pRAM = (uint8*)malloc(RAMSIZE);
	Reset();
}

CGSHandler::~CGSHandler()
{
	FREEPTR(m_pRAM);
}

void CGSHandler::Reset()
{
	memset(m_nReg, 0, sizeof(uint64) * 0x80);
	m_nReg[GS_REG_PRMODECONT] = 1;
	memset(m_pRAM, 0, RAMSIZE);
	m_nPMODE = 0;
}

void CGSHandler::SaveState(CStream* pS)
{
	pS->Write(m_nReg, sizeof(uint64) * 0x80);

	pS->Write(&m_nPMODE,			8);
	pS->Write(&m_nDISPLAY1,			8);
	pS->Write(&m_nDISPFB2,			8);
	pS->Write(&m_nDISPLAY2,			8);
	pS->Write(&m_nCSR,				8);
	pS->Write(&m_nIMR,				8);

	pS->Write(&m_nCrtIsInterlaced,	4);
	pS->Write(&m_nCrtMode,			4);
	pS->Write(&m_nCrtIsFrameMode,	4);

	pS->Write(&m_TrxCtx,			sizeof(TRXCONTEXT));

	pS->Write(m_pRAM,				RAMSIZE);
}

void CGSHandler::LoadState(CStream* pS)
{
	pS->Read(m_nReg, sizeof(uint64) * 0x80);

	pS->Read(&m_nPMODE,				8);
	pS->Read(&m_nDISPLAY1,			8);
	pS->Read(&m_nDISPFB2,			8);
	pS->Read(&m_nDISPLAY2,			8);
	pS->Read(&m_nCSR,				8);
	pS->Read(&m_nIMR,				8);

	pS->Read(&m_nCrtIsInterlaced,	4);
	pS->Read(&m_nCrtMode,			4);
	pS->Read(&m_nCrtIsFrameMode,	4);

	pS->Read(&m_TrxCtx,				sizeof(TRXCONTEXT));

	pS->Read(m_pRAM,				RAMSIZE);

	UpdateViewport();
}

void CGSHandler::SetVBlank()
{
	m_nCSR |= 0x08;

	//Alternate current field
	m_nCSR ^= 0x2000;

	CINTC::AssertLine(CINTC::INTC_LINE_GS);
}

void CGSHandler::ResetVBlank()
{
	m_nCSR &= ~0x08;
}

uint32 CGSHandler::ReadPrivRegister(uint32 nAddress)
{
	uint32 nData;
	switch(nAddress >> 4)
	{
	case 0x1200100:
		//Force CSR to have the H-Blank bit set.
		m_nCSR |= 0x04;
		R_REG(nAddress, nData, m_nCSR);
		break;
	default:
		printf("GS: Read an unhandled priviledged register (0x%0.8X).\r\n", nAddress);
		nData = 0xCCCCCCCC;
		break;
	}
	return nData;
}

void CGSHandler::WritePrivRegister(uint32 nAddress, uint32 nData)
{
	switch(nAddress >> 4)
	{
	case 0x1200000:
		W_REG(nAddress, nData, m_nPMODE);
		if(!(nAddress & 0x4))
		{
			if((m_nPMODE & 0x01) && (m_nPMODE & 0x02))
			{
				printf("GS: Warning. Both read circuits were enabled. Using RC1 for display.\r\n");
				m_nPMODE &= ~0x02;
			}
		}
		break;
	case 0x1200009:
		W_REG(nAddress, nData, m_nDISPFB2);
		if(nAddress & 0x04)
		{
			if(CPS2VM::m_Logging.GetGSLoggingStatus())
			{
				GSDISPFB dispfb;
				DECODE_DISPFB(m_nDISPFB2, dispfb);
				printf("GS: DISPFB2(FBP: 0x%0.8X, FBW: %i, PSM: %i, DBX: %i, DBY: %i);\r\n", \
					dispfb.GetBufPtr(), \
					dispfb.GetBufWidth(), \
					dispfb.nPSM, \
					dispfb.nX, \
					dispfb.nY);
			}
		}
		break;
	case 0x120000A:
		W_REG(nAddress, nData, m_nDISPLAY2);
		if(nAddress & 0x04)
		{
			UpdateViewport();
		}
		break;
	case 0x1200100:
		W_REG(nAddress, nData, m_nCSR);
		if(!(nAddress & 0x04))
		{
			if(nData & 0x08)
			{
				ResetVBlank();
			}
		}
		break;
	case 0x1200101:
		W_REG(nAddress, nData, m_nIMR);
		break;
	default:
		printf("GS: Wrote to an unhandled priviledged register (0x%0.8X, 0x%0.8X).\r\n", nAddress, nData);
		break;
	}
}

void CGSHandler::WriteRegister(uint8 nRegister, uint64 nData)
{
	m_nReg[nRegister] = nData;
	
	switch(nRegister)
	{
	case GS_REG_TRXDIR:
		unsigned int nW, nH, nSX, nSY, nDX, nDY, nDirection, nPixelSize;
		GSBITBLTBUF buf;

		DECODE_TRXREG(m_nReg[GS_REG_TRXREG], nW, nH);
		DECODE_TRXPOS(m_nReg[GS_REG_TRXPOS], nSX, nSY, nDX, nDY, nDirection);
		DECODE_BITBLTBUF(m_nReg[GS_REG_BITBLTBUF], buf);

		nPixelSize = GetPsmPixelSize(buf.nDstPsm);

		m_TrxCtx.nSize		= nW * nH * nPixelSize;
		m_TrxCtx.nCount		= 0;
		m_TrxCtx.nSrcPitch	= nW * nPixelSize;
		m_TrxCtx.nDstPitch	= buf.GetDstWidth() * nPixelSize;
		m_TrxCtx.nPtr		= buf.GetDstPtr() + (nDX * nPixelSize) + (nDY * m_TrxCtx.nDstPitch);

		break;
	}


	DisassembleWrite(nRegister, nData);
}

void CGSHandler::FeedImageData(void* pData, uint32 nLenght)
{
	uint32 nPtr;
	uint32 nLineRest;
	uint32 nLineStat;
	uint32 nSupply;

	if(m_TrxCtx.nSize == 0)
	{
		//printf("GS: Warning. Received image data when no transfer was expected.\r\n");
		return;
	}

	if(nLenght > m_TrxCtx.nSize)
	{
		//assert(0);
		//return;
		nLenght = m_TrxCtx.nSize;
	}

	if(m_TrxCtx.nDstPitch == m_TrxCtx.nSrcPitch)
	{
		memcpy(m_pRAM + m_TrxCtx.nPtr + m_TrxCtx.nCount, pData, nLenght);
		m_TrxCtx.nCount	+= nLenght;
		m_TrxCtx.nSize	-= nLenght;
	}
	else
	{
		nSupply = nLenght;
		while(nSupply != 0)
		{

			//Fill up the remain of the current line
			nLineStat	=  (m_TrxCtx.nCount % m_TrxCtx.nSrcPitch);
			nLineRest	=  m_TrxCtx.nSrcPitch - nLineStat;
			nPtr		=  (m_TrxCtx.nCount / m_TrxCtx.nSrcPitch) * m_TrxCtx.nDstPitch;
			nPtr		+= (m_TrxCtx.nPtr);

			if(nSupply >= nLineRest)
			{
				memcpy(m_pRAM + nPtr + nLineStat, pData, nLineRest);
				*(uint8**)&pData = (uint8*)pData + nLineRest;

				m_TrxCtx.nCount	+= nLineRest;
				m_TrxCtx.nSize	-= nLineRest;

				nSupply -= nLineRest;
			}
			else
			{
				memcpy(m_pRAM + nPtr + nLineStat, pData, nSupply);
				
				m_TrxCtx.nCount	+= nSupply;
				m_TrxCtx.nSize	-= nSupply;

				nSupply = 0;
			}
		}
	}

	if(m_TrxCtx.nSize == 0)
	{
		ProcessImageTransfer(m_TrxCtx.nPtr, m_TrxCtx.nCount);
	}
}

void CGSHandler::SetCrt(bool nIsInterlaced, unsigned int nMode, bool nIsFrameMode)
{
	m_nCrtMode			= nMode;
	m_nCrtIsInterlaced	= nIsInterlaced;
	m_nCrtIsFrameMode	= nIsFrameMode;

	UpdateViewport();
}

unsigned int CGSHandler::GetCrtWidth()
{
	switch(m_nCrtMode)
	{
	case 0x02:
	case 0x03:
	case 0x1C:
		return 640;
		break;
	default:
		assert(0);
		return 640;
		break;
	}
}

unsigned int CGSHandler::GetCrtHeight()
{
	switch(m_nCrtMode)
	{
	case 0x02:
		return 448;
		break;
	case 0x03:
		return 512;
		break;
	case 0x1C:
		return 480;
		break;
	default:
		assert(0);
		return 448;
		break;
	}
}

bool CGSHandler::GetCrtIsInterlaced()
{
	return m_nCrtIsInterlaced;
}

bool CGSHandler::GetCrtIsFrameMode()
{
	return m_nCrtIsFrameMode;
}

unsigned int CGSHandler::GetPsmPixelSize(unsigned int nPSM)
{
	switch(nPSM)
	{
	case GS_PSMCT32:
		return 4;
		break;
	case GS_PSMCT16:
	case GS_PSMCT16S:
		return 2;
		break;
	case GS_PSMT8:
		return 1;
		break;
	default:
		return 4;
		break;
	}
}

void CGSHandler::DisassembleWrite(uint8 nRegister, uint64 nData)
{
	GSPRIM pr;
	double nX, nY, nZ;
	double nU, nV;
	double nS, nT;

	if(!CPS2VM::m_Logging.GetGSLoggingStatus()) return;

	//Filtering
	//if(!((nRegister == GS_REG_FRAME_1) || (nRegister == GS_REG_FRAME_2))) return;
	//if(!((nRegister == GS_REG_TEST_1) || (nRegister == GS_REG_TEST_2))) return;

	switch(nRegister)
	{
	case GS_REG_PRIM:
		DECODE_PRIM(nData, pr);
		printf("GS: PRIM(PRI: %i, IIP: %i, TME: %i, FGE: %i, ABE: %i, AA1: %i, FST: %i, CTXT: %i, FIX: %i);\r\n", \
			pr.nType, \
			pr.nShading, \
			pr.nTexture, \
			pr.nFog, \
			pr.nAlpha, \
			pr.nAntiAliasing, \
			pr.nUseUV, \
			pr.nContext, \
			pr.nUseFloat);
		break;
	case GS_REG_RGBAQ:
		GSRGBAQ rgbaq;
		DECODE_RGBAQ(nData, rgbaq);
		printf("GS: RGBAQ(R: 0x%0.2X, G: 0x%0.2X, B: 0x%0.2X, A: 0x%0.2X, Q: %f);\r\n", \
			rgbaq.nR,
			rgbaq.nG,
			rgbaq.nB,
			rgbaq.nA,
			rgbaq.nQ);
		break;
	case GS_REG_ST:
		DECODE_ST(nData, nS, nT);
		printf("GS: ST(S: %f, T: %f);\r\n", \
			nS, \
			nT);
		break;
	case GS_REG_UV:
		DECODE_UV(nData, nU, nV);
		printf("GS: UV(U: %f, V: %f);\r\n", \
			nU, \
			nV);
		break;
	case GS_REG_XYZ2:
		DECODE_XYZ2(nData, nX, nY, nZ);
		printf("GS: XYZ2(%f, %f, %f);\r\n", \
			nX, \
			nY, \
			nZ);
	case GS_REG_XYZ3:
		DECODE_XYZ2(nData, nX, nY, nZ);
		printf("GS: XYZ3(%f, %f, %f);\r\n", \
			nX, \
			nY, \
			nZ);
		break;
	case GS_REG_TEX0_1:
	case GS_REG_TEX0_2:
		GSTEX0 tex;
		DECODE_TEX0(nData, tex);
		printf("GS: TEX0_%i(TBP: 0x%0.8X, TBW: %i, PSM: %i, TW: %i, TH: %i, TCC: %i, TFX: %i, CBP: 0x%0.8X, CPSM: %i, CSM: %i, CSA: %i, CLD: %i);\r\n", \
			nRegister == GS_REG_TEX0_1 ? 1 : 2, \
			tex.GetBufPtr(), \
			tex.GetBufWidth(), \
			tex.nPsm, \
			tex.GetWidth(), \
			tex.GetHeight(), \
			tex.nColorComp, \
			tex.nFunction, \
			tex.GetCLUTPtr(), \
			tex.nCPSM, \
			tex.nCSM, \
			tex.nCSA, \
			tex.nCLD);
		break;
	case GS_REG_TEX1_1:
	case GS_REG_TEX1_2:
		GSTEX1 tex1;
		DECODE_TEX1(nData, tex1);
		printf("GS: TEX1_%i(LCM: %i, MXL: %i, MMAG: %i, MMIN: %i, MTBA: %i, L: %i, K: %i);\r\n", \
			nRegister == GS_REG_TEX1_1 ? 1 : 2, \
			tex1.nLODMethod, \
			tex1.nMaxMip, \
			tex1.nMagFilter, \
			tex1.nMinFilter, \
			tex1.nMipBaseAddr, \
			tex1.nLODL, \
			tex1.nLODK);
		break;
	case GS_REG_XYOFFSET_1:
	case GS_REG_XYOFFSET_2:
		printf("GS: XYOFFSET_%i(%i, %i);\r\n", \
			nRegister == GS_REG_XYOFFSET_1 ? 1 : 2, \
			(uint32)((nData >>  0) & 0xFFFFFFFF), \
			(uint32)((nData >> 32) & 0xFFFFFFFF));
		break;
	case GS_REG_PRMODECONT:
		printf("GS: PRMODECONT(AC = %i);\r\n", \
			nData & 1);
		break;
	case GS_REG_PRMODE:
		GSPRMODE prm;
		DECODE_PRMODE(nData, prm);
		printf("GS: PMMODE(IIP: %i, TME: %i, FGE: %i, ABE: %i, AA1: %i, FST: %i, CTXT: %i, FIX: %i);\r\n", \
			prm.nShading, \
			prm.nTexture, \
			prm.nFog, \
			prm.nAlpha, \
			prm.nAntiAliasing, \
			prm.nUseUV, \
			prm.nContext, \
			prm.nUseFloat);
		break;
	case GS_REG_TEXFLUSH:
		printf("GS: TEXFLUSH();\r\n");
		break;
	case GS_REG_ALPHA_1:
	case GS_REG_ALPHA_2:
		GSALPHA alpha;
		DECODE_ALPHA(nData, alpha);
		printf("GS: ALPHA_%i(A: %i, B: %i, C: %i, D: %i, FIX: 0x%0.2X);\r\n", \
			nRegister == GS_REG_ALPHA_1 ? 1 : 2, \
			alpha.nA, \
			alpha.nB, \
			alpha.nC, \
			alpha.nD, \
			alpha.nFix);
		break;
	case GS_REG_SCISSOR_1:
		printf("GS: SCISSOR_1(%i, %i, %i, %i);\r\n", \
			(uint32)((nData >>  0) & 0xFFFF), \
			(uint32)((nData >> 16) & 0xFFFF), \
			(uint32)((nData >> 32) & 0xFFFF), \
			(uint32)((nData >> 48) & 0xFFFF));
		break;
	case GS_REG_TEST_1:
	case GS_REG_TEST_2:
		GSTEST tst;
		DECODE_TEST(nData, tst);
		printf("GS: TEST_%i(ATE: %i, ATST: %i, AREF: 0x%0.2X, AFAIL: %i, DATE: %i, DATM: %i, ZTE: %i, ZTST: %i);\r\n", \
			nRegister == GS_REG_TEST_1 ? 1 : 2, \
			tst.nAlphaEnabled, \
			tst.nAlphaMethod, \
			tst.nAlphaRef, \
			tst.nAlphaFail, \
			tst.nDestAlphaEnabled, \
			tst.nDestAlphaMode, \
			tst.nDepthEnabled, \
			tst.nDepthMethod);
		break;
	case GS_REG_FRAME_1:
	case GS_REG_FRAME_2:
		GSFRAME fr;
		DECODE_FRAME(nData, fr);
		printf("GS: FRAME_%i(FBP: 0x%0.8X, FBW: %i, PSM: %i, FBMSK: %i);\r\n", \
			nRegister == GS_REG_FRAME_1 ? 1 : 2, \
			fr.GetBasePtr(), \
			fr.GetWidth(), \
			fr.nPsm, \
			fr.nMask);
		break;
	case GS_REG_ZBUF_1:
	case GS_REG_ZBUF_2:
		GSZBUF zbuf;
		DECODE_ZBUF(nData, zbuf);
		printf("GS: ZBUF_%i(ZBP: 0x%0.8X, PSM: %i, ZMSK: %i);\r\n", \
			nRegister == GS_REG_ZBUF_1 ? 1 : 2, \
			zbuf.GetBasePtr(), \
			zbuf.nPsm, \
			zbuf.nMask);
		break;
	case GS_REG_BITBLTBUF:
		GSBITBLTBUF buf;
		DECODE_BITBLTBUF(nData, buf);
		printf("GS: BITBLTBUF(0x%0.8X, %i, %i, 0x%0.8X, %i, %i);\r\n", \
			buf.GetSrcPtr(), \
			buf.GetSrcWidth(), \
			buf.nSrcPsm, \
			buf.GetDstPtr(), \
			buf.GetDstWidth(), \
			buf.nDstPsm);
		break;
	case GS_REG_TRXPOS:
		printf("GS: TRXPOS(%i, %i, %i, %i, %i);\r\n", \
			(uint32)((nData >>  0) & 0xFFFF), \
			(uint32)((nData >> 16) & 0xFFFF), \
			(uint32)((nData >> 32) & 0xFFFF), \
			(uint32)((nData >> 48) & 0xFF),   \
			(uint32)((nData >> 59) & 0x1));
		break;
	case GS_REG_TRXREG:
		printf("GS: TRXREG(%i, %i);\r\n", \
			(uint32)((nData >>  0) & 0xFFFFFFFF),
			(uint32)((nData >> 32) & 0xFFFFFFFF));
		break;
	case GS_REG_TRXDIR:
		printf("GS: TRXDIR(%i);\r\n", \
			(uint32)((nData >>  0) & 0xFFFFFFFF));
		break;
	default:
		printf("GS: Unknown command (0x%X).\r\n", nRegister); 
		break;
	}
}
